#!/usr/bin/env python
# coding: utf-8
#
#    Author:
#        Mathias Panzenböck <e0427417@student.tuwien.ac.at>
#        http://twoday.tuwien.ac.at/pub/stories/325808/
#
#    Usage:
#        ./graph.py > cpm.dot
#        dot -Tsvg cpm.dot -o cpm.svg
#        inkscape cpm.svg

import sys

data = {
#	   ID: ('Beschreibung (nicht verwendet)', Dauer, Vorgaenger-IDs),
	  '1': ('Aufwandsabsch\xc3\xa4tzung', 3, []),
	  '2': ('Vertrag mit Kunden erstellen', 2, ['1']),
	  '3': ('Projektplan erstellen', 1, ['2']),
	 '4a': ('Anforderungsanalyse (a)', 4, ['3']),
	 '4b': ('Anforderungsanalyse (b)', 2, ['4a']),
	 '4c': ('Anforderungsanalyse (c)', 4, ['4b']),
	  '5': ('Anwendungsfallbeschreibung', 5, ['4a']),
	  '6': ('Software Architekturbeschreibung', 6, ['4b']),
	  '7': ('Domainenmodell', 3, ['6']),
	  '8': ('Bereitstellen eines Versionierungssystems', 2, ['3']),
	  '9': ('Bereitstellen eines Configuration Management Systems', 2, ['3']),
	 '10': ('Bereitstellen eines Bugtrackers', 2, ['3']),
	 '11': ('Designdokument', 10, ['4a']),
	 '12': ('Datenmodell erstellen', 2, ['4c', '5', '7']),
	 '13': ('Datenmodell umsetzen', 2, ['12']),
	 '14': ('DB-Handler entwerfen', 2, ['12']),
	 '15': ('DB-Handler implementieren', 4, ['13', '14']),
	 '16': ('DB-Handler testen (Lasttest f\xc3\xbcr Connectionpool)', 3, ['15']),
	 '17': ('Testdaten einpflegen', 2, ['16']),
	 '18': ('Corporate Design (Logo, Farben, etc)', 5, ['3']),
	 '19': ('Usability Guidelines', 2, ['18']),
	 '20': ('UI Prototyp (Frontend Backend)', 3, ['19']),
	 '21': ('User-Testing', 2, ['7', '20']),
	 '22': ('Redesign anhand der Ergebnisse', 2, ['21']),
	 '23': ('Umsetzung der Templates', 10, ['22']),
	 '24': ('Testplan erstellen', 1, ['11']),
	 '25': ('Testf\xc3\xa4lle erstellen', 3, ['24']),
	 '26': ('Laufende Tests', 22, ['16', '25']),
	 '27': ('Integrationstests', 5, ['26', '34']),
	 '28': ('Auktionshandler', 17, ['16', '25']),
	 '29': ('Such-Handler', 12, ['16', '25']),
	 '30': ('User-Handler', 12, ['16', '25']),
	 '31': ('Useradmin', 10, ['16', '25']),
	 '32': ('Auto-Handler', 14, ['16', '25']),
	 '33': ('Autoadmin', 10, ['16', '25']),
	 '34': ('Anbindung der Templates', 5, ['23', '28', '29', '30', '31', '32', '33']),
	 '35': ('Neue Serverhardware bereitstellen', 5, ['11']),
	 '36': ('DB am Server installieren', 3, ['16', '35']),
	 '37': ('DB am Server testen', 2, ['36']),
	 '38': ('Installation der Applikation', 2, ['27', '37']),
	 '39': ('Testen des Live Systems', 3, ['38']),
	 '40': ('Anwender Schulung', 5, ['39']),
	'41a': ('Pause', 5, ['40']),
	'41b': ('Akzeptanztests', 2, ['41a']),
}

def find_successors(data,nr):
	L=[]
	for succ_nr, item in data.iteritems():
		for dep in item[2]:
			if dep == nr:
				L.append(succ_nr)
				break
	return L

class Node(object):
	__slots__='nr','earliest_end','latest_start','dummy', 'preds','succs','critical'
	def __init__(self,nr,dummy=False):
		self.nr=nr
		self.preds=set()
		self.succs=set()
		self.earliest_end=None
		self.latest_start=None
		self.dummy=dummy
		self.critical=False
	
	def __hash__(self):
		return self.nr
	
	def __eq__(self,other):
		return self.nr == other.nr
	
	def __repr__(self):
		return 'Node(nr=%r,earliest_end=%r,latest_start=%r,dummy=%r)' % (
			self.nr,self.earliest_end,self.latest_start,self.dummy)
	
	def has_open_prec_ends(self):
		for edge in self.preds:
			if not edge.visited_earliest_end:
				return True
		return False

	def has_open_succ_starts(self):
		for edge in self.succs:
			if not edge.visited_latest_start:
				return True
		return False

	def has_open_relabel(self):
		for edge in self.preds:
			if not edge.visited_relabel:
				return True
		return False

#	def relabel_dummy_edges(self):
#		dummy_nr = 1
#		edges = list(self.succs)
#		while edges:
#			edge = edges.pop()
#			if edge.dummy:
#				edge.label = 'S%d' % dummy_nr
#				dummy_nr += 1
#			edge.visited_relabel = True
#			if not edge.target.has_open_relabel():
#				edges.extend(edge.target.succs)

	def relabel_dummy_edges(self):
		dummy_nr = 1
		edges = list(self.succs)
		while edges:
			edge = edges.pop()
			if not edge.visited_relabel:
				if edge.dummy:
					edge.label = 'S%d' % dummy_nr
					dummy_nr += 1
				edge.visited_relabel = True
				oldedges = edges
				edges = [e for e in edge.target.succs if not e.visited_relabel]
				edges.extend(oldedges)

	def calc_earlisest_end(self):
		self.earliest_end = 0
		edges = list(self.succs)
		
		while edges:
			edge = edges.pop()
			edge.visited_earliest_end = True
			node = edge.target
			new_time = edge.source.earliest_end + edge.weight
			if node.earliest_end is None or node.earliest_end < new_time:
				node.earliest_end = new_time
			
			if not node.has_open_prec_ends():
				edges.extend(node.succs)
	
	def calc_latest_start(self):
		self.latest_start = self.earliest_end
		edges = list(self.preds)
		
		while edges:
			edge = edges.pop()
			edge.visited_latest_start = True
			node = edge.source
			new_time = edge.target.latest_start - edge.weight
			if node.latest_start is None or node.latest_start > new_time:
				node.latest_start = new_time

			if not node.has_open_succ_starts():
				edges.extend(node.preds)
	
	def mark_critical(self):
		self.critical = True

		for edge in self.preds:
			if edge.source.earliest_end + edge.weight == self.earliest_end:
				edge.critical = True
				edge.source.mark_critical()

class Edge(object):
	__slots__='nr','source','target','weight','dummy','critical','visited_earliest_end','visited_latest_start','visited_relabel','label'
	def __init__(self,nr,source,target,weight=0,dummy=False):
		self.nr=nr
		self.source=source
		self.target=target
		self.weight=weight
		self.dummy=dummy
		self.critical=False
		self.visited_earliest_end = False
		self.visited_latest_start = False
		self.visited_relabel      = False

		if dummy:
			self.label=str(nr)
		else:
			self.label='%s, %dt' % (nr, weight)

	def __hash__(self):
		return hash(self.nr)
	
	def __eq__(self,other):
		return self.nr == other.nr
	
	def __repr__(self):
		return 'Edge(nr=%r,source=%r,target=%r,weight=%r,dummy=%r)' % (
			self.nr, self.source.nr, self.target.nr, self.weight, self.dummy)

class NodeSet(dict):
	def create(self,dummy=False):
		node = Node(len(self)+1,dummy)
		assert self.get(node.nr,None) is None
		self[node.nr] = node
		return node

class EdgeSet(dict):
	__slots__='next_dummy_id',
	def __init__(self,*args,**kwargs):
		dict.__init__(self,*args,**kwargs)
		self.next_dummy_id = 0

def make_graph(data,fout):
	edges  = {}
	nodes  = NodeSet()

	def new_dummy_edge(xedge,enr,target):
		if xedge.target is None:
			build_simple(nodes.create(dummy=True),edges[enr],xedge)
		else:
			assert xedge.target.dummy

		source = xedge.target
		edge = Edge('dummy_%d' % xedges.next_dummy_id,source,target,dummy=True)
		xedges.next_dummy_id += 1
		source.succs.add(edge)
		target.preds.add(edge)
		assert edge.nr not in xedges
		xedges[edge.nr] = edge
		return edge
	
	def build_simple(node,edge,xedge):
		xedge.target = node
		node.preds.add(xedge)
		if xedge.source is None:
			xedge.source = nodes.create()
			xedge.source.succs.add(xedge)
			build_graph(xedge.source, edge[2])

	def build_graph(node,pred_edges):
		for enr in pred_edges:
			xedge = xedges[enr]
			edge  = edges[enr]
			succs = edge[3]
			if len(succs) > 1:
				# get/create dummy node and create dummy edge
				new_dummy_edge(xedge,enr,node)
			else:
				build_simple(node,edge,xedge)

	for nr, (label, duration, deps) in data.iteritems():
		edges[nr] = (label, duration, deps, find_successors(data,nr))
	
	xedges = EdgeSet((enr, Edge(enr,None,None,edata[1])) for enr, edata in edges.iteritems())
	end_edges = [enr for enr, edata in edges.iteritems() if not edata[3]]
	end_node = nodes.create()
	build_graph(end_node,end_edges)
	
	start_nodes = [node for node in nodes.itervalues() if not node.preds]
	start_node = start_nodes[0]

	# reduce start nodes:
	for node in start_nodes[1:]:
		for edge in node.succs:
			edge.source = start_node
			start_node.succs.add(edge)
		del nodes[node.nr]

	# optimize graph:
	for node in nodes.values():
		if len(node.preds) == 1 and len(node.succs) == 1:
			pred_edge = iter(node.preds).next()
			succ_edge = iter(node.succs).next()
			pred = pred_edge.source
			succ = succ_edge.target

			if not pred_edge.dummy:
				continue

			can_opt = True
			# can only opt if there is no sibling with same pred and succ
			for sibling_edge in pred.succs:
				if sibling_edge.target is succ:
					can_opt = False
					break

			if can_opt:
				del xedges[pred_edge.nr]
				del nodes[node.nr]
				succ_edge.source = pred
				pred.succs.remove(pred_edge)
				pred.succs.add(succ_edge)

	# calculate times:
	start_node.calc_earlisest_end()
	end_node.calc_latest_start()
	end_node.mark_critical()

	# rename dummies (in order!) because the optimisations might have deleted some names
	start_node.relabel_dummy_edges()

	# generate dot graph:
#	fout.write('digraph CPM {\n\trank=LR\n\tnodesep=0.6\n\tnode[shape=Mrecord];\n\n')
	fout.write('digraph CPM {\n\tnode[shape=Mrecord];\n\n')

	for node in nodes.itervalues():
		label = "{%s|%s}" % (node.earliest_end, node.latest_start)

		if not node.preds or not node.succs:
			label = "|%s|" % label

		fout.write('\tN_%s [label="%s"' % (node.nr, label))
		if node.critical:
			fout.write(',color=red,fontcolor=red')
		fout.write(']\n')

	fout.write('\n')

	for edge in xedges.itervalues():
		if edge.dummy:
			fout.write('\tN_%s->N_%s [label="%s",style=dashed' % (edge.source.nr, edge.target.nr, edge.label))
		else:
			fout.write('\tN_%s->N_%s [label="%s"' % (edge.source.nr, edge.target.nr, edge.label))
		if edge.critical:
			fout.write(',color=red,fontcolor=red')
		fout.write(']\n')

	fout.write('}\n')

if __name__ == '__main__':
	make_graph(data,sys.stdout)
